# coding:utf-8
'''
表格生成线条坐标
'''
import cv2
import numpy as np
import os
import pytesseract
import uuid

# 图像切割模块


class cutImage(object):
    def __init__(self, img, **settings):
        '''
        :param img: 输入图像
        :param bin_threshold: 二值化的阈值大小
        :param kernel: 形态学kernel
        :param iterations: 迭代次数
        :param areaRange: 面积范围 单元格的面积大小 此参数决定哪些是单元格，例如:[9000, 100000]
        :param border: 留边大小
        :param show: 是否显示结果图，默认是显示
        :param write: 是否把结果写到文件，默认是写入
        :param lang:ocr识别的语言，例如：'eng+tha'
        '''
        self.img = img

        self.bin_threshold = settings.get('pytesseract_bin_threshold', 127)
        self.kernel = settings.get(
            'pytesseract_kernel', np.ones((4, 4), np.uint8))
        self.iterations = settings.get('pytesseract_iterations', 1)
        self.areaRange = settings.get('pytesseract_areaRange', [10000, 100000])
        self.border = settings.get('pytesseract_border', 10)
        self.isDebug = settings.get('pytesseract_isDebug', False)
        self.lang = settings.get('pytesseract_lang', 'eng')

    def hitTest(self, p, x, y, w, h):
        '''
        : 得到包含在面里的点数组元素
        ：x y w h构成一个面
        '''
        x1 = x
        x2 = x + w
        y1 = y
        y2 = y + h

        px = p['x'] + p['w'] / 2
        py = p['y'] + p['h'] / 2

        return (px >= x1 and px <= x2) and (py >= y1 and py <= y2)

    def draw_points(self, img, pointArr):
        # 在坐标点上画点
        for point in pointArr:
            cv2.circle(
                img, (point['x'] + int(point['w'] / 2), point['y'] + int(point['h'] / 2)), 1, (0, 0, 255), 10)

        # cv2.imshow('point', img)
        cv2.imwrite('bitwiseAnd_point.png', img)
        # cv2.waitKey()

    def get_text(self):
        # 得到mask图和网格图
        mask, joint = detect_table(self.img)
        # 得到交叉点坐标
        pointArr = find_joint_points(joint)
        if pointArr is None:
            return None

        if self.img.shape[2] == 1:  # 灰度图
            img_gray = self.img
        elif self.img.shape[2] == 3:
            img_gray = cv2.cvtColor(self.img, cv2.COLOR_BGR2GRAY)
        ret, thresh = cv2.threshold(
            img_gray, self.bin_threshold, 255, cv2.THRESH_BINARY_INV)  # 二值化
        img_erode = cv2.dilate(thresh, self.kernel, iterations=self.iterations)

        contours, hierarchy = cv2.findContours(
            img_erode, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)

        result = {}
        area_coord_roi = []

        npPointArr = np.array(pointArr)
        row, col = npPointArr.shape
        data = np.empty((row, col), dtype=object)

        for i in range(len(contours)):
            cnt = contours[i]
            area = cv2.contourArea(cnt)
            if area > self.areaRange[0] and area < self.areaRange[1]:
                x, y, w, h = cv2.boundingRect(cnt)
                roi = self.img[y+self.border:(y+h)-self.border,
                               x+self.border:(x+w)-self.border]
                area_coord_roi.append((area, (x, y, w, h), roi))

                text = None
                if self.isDebug == False:
                    text = pytesseract.image_to_string(
                        roi, lang=self.lang, config='--dpi 300')  # 读取文字
                    print('tesseract cell text:', text)

                if self.isDebug:
                    cv2.rectangle(
                        self.img, (x, y), (x + w, y + h), (0, 255, 0), 2)

                # cv2.imshow('cut_img', roi)
                # cv2.waitKey()

                # 根据坐标，将表格值写入数组中
                for row in pointArr:
                    pList = [p for p in row if self.hitTest(p, x, y, w, h)]
                    if len(pList) > 0:
                        firtP = pList[0]
                        cx = firtP['arrX']
                        cy = firtP['arrY']
                        data[cx][cy] = text

                        # None为''
                        for p in pList[1:]:
                            cx = p['arrX']
                            cy = p['arrY']
                            data[cx][cy] = ''

                        # 画点用来debug
                        if self.isDebug:
                            for point in pList:
                                cv2.circle(
                                    self.img, (point['x'] + int(point['w'] / 2), point['y'] + int(point['h'] / 2)), 1, (0, 0, 255), 10)

        if self.isDebug:
            cv2.imshow("marked-image-debug", self.img)
            cv2.imwrite("marked-image-debug.png", self.img)
            cv2.waitKey()

        return data.tolist()


def detect_table(img):
    '''
    : 探测表格，得到网格图和网格坐标点图
    : parame: src_img 原始图像
    : 返回：
    :   mask_img 网格图，由横线和竖线交叉得到的网格图片
    :   joints_img 坐标点图，网格的交点图片
    '''
    if len(img.shape) == 2:  # 灰度图
        gray_img = img
    elif len(img.shape) == 3:
        gray_img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)

    thresh_img = cv2.adaptiveThreshold(
        ~gray_img, 255, cv2.ADAPTIVE_THRESH_MEAN_C, cv2.THRESH_BINARY, 15, -2)
    h_img = thresh_img.copy()
    v_img = thresh_img.copy()
    scale = 15
    h_size = int(h_img.shape[1]/scale)

    h_structure = cv2.getStructuringElement(
        cv2.MORPH_RECT, (h_size, 1))  # 形态学因子
    h_erode_img = cv2.erode(h_img, h_structure, 1)

    h_dilate_img = cv2.dilate(h_erode_img, h_structure, 1)
    # cv2.imshow("h_erode",h_dilate_img)
    v_size = int(v_img.shape[0] / scale)

    v_structure = cv2.getStructuringElement(
        cv2.MORPH_RECT, (1, v_size))  # 形态学因子
    v_erode_img = cv2.erode(v_img, v_structure, 1)
    v_dilate_img = cv2.dilate(v_erode_img, v_structure, 1)

    mask_img = h_dilate_img+v_dilate_img
    joints_img = cv2.bitwise_and(h_dilate_img, v_dilate_img)

    # cv2.imshow("joints", joints_img)
    # cv2.imshow("mask", mask_img)
    return mask_img, joints_img


def find_table(img, mask_img, save=False, save_dir=None):
    '''
    : 查找表格
    : parame: img 图片
    : parame: contours
    : 返回 表格数组:
    : x 表格左上顶点的x值
    : y 表格坐上顶点的y值
    : w 表格宽度
    : h 表格高度
    '''
    contours, hierarchy = cv2.findContours(
        mask_img, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
    contours = sorted(contours, key=cv2.contourArea, reverse=True)
    tables = []
    for i in range(len(contours)):
        area0 = cv2.contourArea(contours[i])
        if area0 < 20:
            continue

        # =======================查找每个表的关节数====================
        epsilon = 0.1 * cv2.arcLength(contours[i], True)
        approx = cv2.approxPolyDP(contours[i], epsilon, True)  # 获取近似轮廓
        x, y, w, h = cv2.boundingRect(approx)

        table_path = None
        # 保存文件
        if save == True:
            table_img = img[y: y + h, x: x + w]
            if save_dir is None:
                save_dir = os.getcwd() + '/tmp/tables'
            # 临时文件夹不存在则创建
            if os.path.exists(save_dir) == False:
                os.makedirs(save_dir)
            uuid_id = uuid.uuid1()
            table_path = save_dir + '/' + str(uuid_id) + '.png'

            cv2.imwrite(table_path, table_img)

        tables.append({'x': x, 'y': y, 'w': w, 'h': h,
                       'table_path': table_path})

    return tables


def find_joint_points(joint):
    '''
    : 根据joint得到网格图信息
    ：parame: joint 坐标点图片数据
    ：返回：
    ：pointArr：包含坐标点信息的数组
    ：      arrX: 数组中的行索引
    :       arrY: 数组中的列索引
    :       x: x坐标
    :       y: y坐标
    :       w: 宽度
    :       h: 高度
    '''
    ys, xs = np.where(joint > 0)

    if len(xs) == 0:
        return None

    mylisty = []  # 纵坐标
    mylistx = []  # 横坐标

    # 通过排序，获取跳变的x和y的值，说明是交点，否则交点会有好多像素值值相近，我只取相近值的最后一点
    # 这个10的跳变不是固定的，根据不同的图片会有微调，基本上为单元格表格的高度（y坐标跳变）和长度（x坐标跳变）
    i = 0
    myxs = np.sort(xs)
    for i in range(len(myxs)-1):
        if(myxs[i+1]-myxs[i] > 10):
            mylistx.append(myxs[i])
        i = i+1
    mylistx.append(myxs[i])  # 要将最后一个点加入

    i = 0
    myys = np.sort(ys)
    # print(np.sort(ys))
    for i in range(len(myys)-1):
        if(myys[i+1]-myys[i] > 10):
            mylisty.append(myys[i])
        i = i+1
    mylisty.append(myys[i])  # 要将最后一个点加入

    pointArr = []

    # 循环y坐标，x坐标分割表格
    for i in range(len(mylisty) - 1):
        rowArr = []

        for j in range(len(mylistx) - 1):

            x1 = mylistx[j]
            x2 = mylistx[j + 1]
            y1 = mylisty[i]
            y2 = mylisty[i + 1]

            rowArr.append({'arrX': i, 'arrY': j, "x": x1, "y": y1,
                           'w': x2 - x1, 'h': y2-y1})
            j = j + 1

        pointArr.append(rowArr)
        i = i+1
    return pointArr


def extract_tables(img_path, **settings):
    '''
    抽取表格
    '''
    # 读取图片
    img = cv2.imread(img_path)

    mask, joints = detect_table(img)

    # 寻找表格并且保存为图片
    table_images = find_table(img, mask, save=True)

    # 对表格图片进行分割并使用ocr得到文本
    kernel = np.ones((4, 4), np.uint8)
    tables = []

    # 循环表格抓取数据
    for tImg in table_images:
        table_path = tImg['table_path']

        # 切图并识别成表格
        table_img = cv2.imread(table_path)
        data = cutImage(table_img, **settings).get_text()

        if data is not None:
            tables.append(data)

        # 删除临时文件
        os.remove(table_path)

    return tables


if __name__ == '__main__':
    img_path = 'C:/Work/HxProjects/Wpbs/Crawler/pdf2tables/test_data/Jan-2010-page-2-300.png'

    isDebug = False
    # 切图并识别成表格
    settings = {
        'pytesseract_kernel': np.ones((4, 4), np.uint8),
        'pytesseract_bin_threshold': 127,
        'pytesseract_iterations': 1,
        'pytesseract_areaRange': [10000, 100000],
        'pytesseract_isDebug': isDebug,
        'pytesseract_border': 10
    }

    tables = extract_tables(
        img_path, **settings)

    print(len(tables))

    if isDebug == False:
        for data in tables:
            for row in data:
                print(row)
            print('*********************************************************************')




import sys
import os
import base64
import time
import json

from urllib.parse import urlparse
import traceback
from urllib import request
import base64


class AliyunOcrTableParse:
    def get_img_base64_from_file(self, img_file):
        with open(img_file, 'rb') as infile:
            s = infile.read()
            return self.get_img_base64(s)

    def get_img_base64(self, img_byte):
        byte64 = base64.b64encode(img_byte)
        return str(byte64, encoding="utf-8")

    def predict(self, url, appcode, img_base64, kv_config, old_format):
        if not old_format:
            param = {}
            param['image'] = img_base64
            if kv_config is not None:
                param['configure'] = json.dumps(kv_config)
            body = json.dumps(param)
        else:
            param = {}
            pic = {}
            pic['dataType'] = 50
            pic['dataValue'] = img_base64
            param['image'] = pic

            if kv_config is not None:
                conf = {}
                conf['dataType'] = 50
                conf['dataValue'] = json.dumps(kv_config)
                param['configure'] = conf

            inputs = {"inputs": [param]}
            body = json.dumps(inputs)

        bodyBytes = bytes(body, encoding="utf8")

        headers = {'Authorization': 'APPCODE %s' % appcode}
        requestOjb = request.Request(url=url, headers=headers, data=bodyBytes)
        try:
            response = request.urlopen(requestOjb, timeout=10)
            return response.code, response.headers, response.read()
        except request.HTTPError as e:
            return e.code, e.headers, e.read()

    def parse(self, appcode, img_byte, format='json'):
        url = 'https://form.market.alicloudapi.com/api/predict/ocr_table_parse'
        img_base64data = self.get_img_base64(img_byte)
        # 如果输入带有inputs, 设置为True，否则设为False
        is_old_format = False
        config = {'format': format, 'finance': False, 'dir_assure': False}
        # 如果没有configure字段，config设为None
        # config = None
        stat, header, content = self.predict(
            url, appcode, img_base64data, config, is_old_format)

        contentStr = str(content, 'utf-8')

        if stat != 200:
            err = 'Http status code: ' + str(stat)
            err = err + 'Error msg in header: ' + \
                header['x-ca-error-message'] if 'x-ca-error-message' in header else ''
            err = err + 'Error msg in body: ' + contentStr
            raise Exception(err)
        if is_old_format:
            result_str = json.loads(
                contentStr)['outputs'][0]['outputValue']['dataValue']
        else:
            result_str = contentStr

        return result_str

    # appcode   应用的AppCode
    # img_file  图片路径
    # format    输出的格式 html/json/xlsx
    def extract_tables(self, img_file, **settings):
        appcode = settings.get('aliyun_appcode')

        if appcode is None:
            raise Exception('Aliyun Appcode required')

        format = 'json'
        datas = []

        with open(img_file, 'rb') as infile:
            s = infile.read()
            json_str = self.parse(appcode, s, format)

            json_data = json.loads(json_str)
            # 转化json为list
            for table in json_data['tables']:
                data = []
                for row in table:
                    values = []
                    for col in row:
                        if type(col) is dict:
                            if col['text'] is not None:
                                values.append(col['text'][0])
                    if len(values) > 0:
                        data.append(values)
                datas.append(data)

        return datas


def extract_tables(img_path, **settings):
    print('extract tables using aliyun ocr...')
    datas = AliyunOcrTableParse().extract_tables(img_path, **settings)
    return datas


if __name__ == "__main__":
    img_path = 'C:/Work/HxProjects/Wpbs/Crawler/pdf2tables/test_data/Jan-2010-page-2-300.png'
    datas = extract_tables(
        img_path, aliyun_appcode='a8f41a5f9b664a45af2bc9f58666a17e')

    for data in datas:
        print('*****************************************')
        for row in data:
            print(row)



from setuptools import find_packages, setup
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='pdf2tables',
    version='0.3.1.4',
    packages=find_packages(where='.', exclude=(), include=('*',)),
    url='https://github.com/chen1tian/pdf2tables',
    license='BSD 2-Clause License',
    author='yitian.chen',
    author_email='48295852@qq.com',
    description='extract tables from pdf using camelot, if page is image-base, use ocr to extract',
    long_description_content_type='text/markdown',
    long_description=long_description,
    install_requires=[
        'opencv-python >= 4.0.0.21',
        'pytesseract >= 0.2.6',
        'camelot-py >= 0.7.3',
        'numpy >= 1.16.0'
    ]
)



import os
from dataclasses import dataclass

import camelot
import numpy as np
import pdfplumber
import pytesseract
from PIL import Image
from enum import Enum


@dataclass
class PageTable:
    '''
    表格抽取结果类
    '''

    # 页数
    page: int
    # 数据列表
    datas: []
    # 本页文本
    text: str


class ImgOcrType(Enum):
    '''
    图片Ocr类型
    '''

    # 阿里云
    Pytesseract = 1
    Aliyun = 2


def get_page_img_path(pdf: pdfplumber.PDF, page_number: str):
    '''
    : 得到pdf页面转换为图片的路径
    '''
    page_num_str = str(page_number)
    pdf_path = pdf.stream.name
    basename = os.path.basename(pdf_path)
    root_name, ext = os.path.splitext(basename)
    img_path = os.getcwd() + '/tmp/' + root_name + \
        '-page-' + page_num_str + '.png'

    img_path = img_path.replace('\\\\', '/')

    # 目录不存在则建立
    img_dir = os.path.dirname(img_path)
    if os.path.exists(img_dir) == False:
        os.makedirs(img_dir)

    return img_path


def merge_table(all_tables: [PageTable], page_number: int, data, text: str):
    '''
    : 合并表格数据
    '''
    same_tables = [
        table for table in all_tables if table.page == page_number]
    if same_tables:
        same_table = same_tables[0]
        same_table.datas.append(data)
        same_table.text = text
    else:
        all_tables.insert(
            page_number - 1, PageTable(page_number, [data], text))


def merge_tables(all_tables, page_number, other_tables):
    for t in other_tables:
        for data in t.datas:
            merge_table(all_tables, page_number, data, t.text)


def extract_imgbase(pdf: pdfplumber.PDF, page_number: int, flavor: str, lang: str, filter=None, **imgOcrSettings):
    '''
    : 抽取基于图片的pdf
    '''
    print('use ocr extract page {page_number} tables'.format(
        page_number=page_number))

    # 保存图片
    page_img_path = get_page_img_path(pdf, page_number)

    page = pdf.pages[page_number - 1]
    page_img = page.to_image(resolution=300)
    page_img.save(page_img_path, format='PNG')

    page_text = pytesseract.image_to_string(
        page_img_path, lang=lang, config='--psm 11 --dpi 300', nice=1)

    if filter is not None:
        needExtract = filter(page_text, page_number, page_img_path)
        if needExtract == False:
            print('不符合过滤条件，略过')
            return None

    # 处理配置 根据ocr类型，删除无关的配置项
    img_ocr_type = imgOcrSettings.get(
        'img_ocr_type', ImgOcrType.Pytesseract)

    ocr_type_name = img_ocr_type.name

    setting_keys = [k for k in imgOcrSettings.keys(
    ) if k[0: len(ocr_type_name)].lower() != ocr_type_name.lower()]

    for key in setting_keys:
        del imgOcrSettings[key]

    # 得到图片列表处理者
    imgtable = None
    if img_ocr_type == ImgOcrType.Pytesseract:
        import pdf2tables.image_tables as imgtable

    if img_ocr_type == ImgOcrType.Aliyun:
        import pdf2tables.aliyun_tables as imgtable

    datas = imgtable.extract_tables(page_img_path, **imgOcrSettings)

    tables = []
    tables.append(PageTable(page_number, datas, page_text))

    # 删除临时的pdf页面图片
    os.remove(page_img_path)

    return tables


def extract(pdf_path: str, filter=None, flavor='lattice', lang: str = 'eng', **imgOcrSettings):
    '''
    : 抽取pdf中的表格数据
    '''
    pdf = pdfplumber.from_path(pdf_path)
    total_page = len(pdf.pages)

    tables: [PageTable] = []

    # 使用camelot抽取表格
    print('use camelot extract tables')
    camelot_tables = camelot.read_pdf(pdf_path, pages='all',
                                      flavor=flavor, suppress_stdout=False)

    for t in camelot_tables:
        text = pdf.pages[t.page - 1].extract_text()
        merge_table(tables, t.page, t.data, text)

    # 如果抽取完成则返回
    if len(tables) == total_page:
        return tables

    # 否则使用ocr抽取其他页面的表格
    extract_pages = [t.page for t in tables]

    total_page_set = set(range(1, total_page + 1))
    extracted_pages_set = set(extract_pages)

    other_pages = list(total_page_set.difference(extracted_pages_set))

    for page_number in other_pages:
        other_tables = extract_imgbase(
            pdf, page_number, flavor, lang, filter, **imgOcrSettings)

        if other_tables is not None:
            merge_tables(tables, page_number, other_tables)

    pdf.close()

    return tables


def test_filter(page_text, page_number, page_img_path):
    text_arr = [t for t in page_text.split('\n',) if t]
    return text_arr[0].upper() == 'Report of Sugar Export'.upper() and text_arr[1].upper() == 'JANUARY 2010'.upper()


if __name__ == '__main__':

    imgOcrSettings = {
        'pytesseract_kernel': np.ones((4, 4), np.uint8),
        'pytesseract_bin_threshold': 127,
        'pytesseract_iterations': 1,
        'pytesseract_areaRange': [10000, 100000],
        'pytesseract_isDebug': False,
        'pytesseract_border': 10,
        'img_ocr_type': ImgOcrType.Pytesseract,
        'aliyun_appcode': 'a8f41a5f9b664a45af2bc9f58666a17e'
    }

    tables = extract(
        'C:/Work/HxProjects/Wpbs/Crawler/pdf2tables/test_data/Jan-2010.pdf', lang='eng+tha', filter=test_filter, ** imgOcrSettings)

    for t in tables:
        print("page", t.page)
        for i in range(len(t.datas)):
            print('data' + str(i),
                  '************************************************************')
            data = t.datas[i]
            for row in data:
                print(row)
